"""Main bitrix24 module."""
import asyncio
import time

import httpx

from aiobitrix24 import _builders
from aiobitrix24.methods.common import BATCH


class _Bitrix24:
    """Class for making bitrix24 requests."""

    def __init__(
        self,
        url: str | None = None,
        sleep_sec: int = 3,
        httpx_timeout: int = 20,
    ) -> None:
        """Init class.

        :param url: base url for bitrix24 rest api, defaults to None
        :param sleep_sec: sleep seconds between queries, defaults to 3
        :param httpx_timeout: request timeout, defaults to 20
        """
        self._url = url
        self._sleep_sec = sleep_sec
        self._httpx_timeout = httpx_timeout
        self._last_request_time = time.time() - sleep_sec

    def url(self, url: str) -> None:
        """Set base url for Birtix24 class.

        :param url: base bitrix url
        """
        self._url = url

    def timeout(self, timeout: int | float) -> None:
        """Set bitrix timeout.

        :param timeout: timeout_sec
        """
        self._sleep_sec = timeout

    async def request(
        self,
        method: str,
        query_params: dict,
        is_query_complex: bool = False,
    ) -> httpx.Response:
        """Request to bitrix24.

        :param method: call method
        :param query_params: query params
        :param is_query_complex: build custom param string, defaults to False
        :raises ValueError: url parameter not set
        :return: http response
        """
        if self._url is None:
            raise ValueError("Url parameter not set")
        await self._sleep()
        async with httpx.AsyncClient() as client:
            if is_query_complex:
                complex_query = _builders.build_query(query_params)
                return await client.post(
                    f"{self._url}{method}?{complex_query}",
                    timeout=self._httpx_timeout,
                )
            return await client.post(
                f"{self._url}{method}",
                json=query_params,
                timeout=self._httpx_timeout,
            )

    async def batch_request(
        self,
        queries: list[_builders.BatchQuery],
    ) -> httpx.Response:
        """Batch request to bitrix24.

        :param queries: batch of queries
        :raises ValueError: url parameter not set
        :return: http response
        """
        if self._url is None:
            raise ValueError("Url parameter not set")
        await self._sleep()
        batch_query = _builders.build_batch(queries)
        async with httpx.AsyncClient() as client:
            return await client.post(
                f"{self._url}{BATCH}",
                json=batch_query,
                timeout=self._httpx_timeout,
            )

    async def _sleep(self) -> None:
        current_time = time.time()
        if current_time - self._last_request_time < self._sleep_sec:
            self._last_request_time = current_time + self._sleep_sec
            await asyncio.sleep(self._sleep_sec)
        else:
            self._last_request_time = current_time


bx24 = _Bitrix24()



"""Module exceptions."""


class BitrixError(Exception):
    """Main bitrix error."""


class BadBitrixResponseError(BitrixError):
    """Bad response from bx24."""


class BatchError(BitrixError):
    """Rises if batch size is more than 50."""



from __future__ import annotations

import logging
from dataclasses import dataclass

from aiobitrix24._bitrix24 import bx24
from aiobitrix24._builders import MAX_BATCH_SIZE, BatchQuery, build_chunks
from aiobitrix24.exceptions import BitrixError
from aiobitrix24.methods import crm

logger = logging.getLogger("aiobitrix24.crm")
logger.setLevel(logging.WARNING)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)

logger.addHandler(ch)


class BaseSelect:
    @classmethod
    async def json(
        cls,
        method: str,
        start: int,
        process_id: int,
        filters: dict,
        select: list[str] | None = None,
    ) -> dict:
        query_params = {
            "start": start,
            "entityTypeId": process_id,
            "select": select if select else ["*"],
            "filter": filters,
        }
        response_json = (
            await bx24.request(
                method,
                query_params,
            )
        ).json()
        if "error" in response_json:
            raise BitrixError(response_json["error_description"])
        return response_json


# class BaseBatchUpdate:
#     @classmethod
#     async def update(chunk)


@dataclass
class CRMItem:
    """Item in bitrix crm smart process."""

    fields: dict
    name: str = ""
    _select = BaseSelect

    @classmethod
    async def select(
        cls,
        process_id: int,
        filters: dict,
        select: list[str] | None = None,
    ) -> list[CRMItem]:
        """Select all items from process satisfying filters.

        :param process_id: smart process id in bitrix
        :param filters: filters like in crm docs
        :param select: field for select, defaults to None
        :return: list of selected smart process items
        """
        response_json = await cls._select.json(
            crm.Item.LIST,
            0,
            process_id,
            filters,
            select,
        )
        logger.debug(response_json)
        crm_items = [CRMItem(fields) for fields in response_json["result"]["items"]]
        for page in range(MAX_BATCH_SIZE, response_json["total"], MAX_BATCH_SIZE):
            response_json = await cls._select.json(
                crm.Item.LIST,
                page,
                process_id,
                filters,
                select,
            )
            page_items = [
                CRMItem(fields) for fields in response_json["result"]["items"]
            ]
            crm_items.extend(page_items)
        return crm_items

    @classmethod
    async def batch_update(cls, crm_items: list[CRMItem]) -> None:
        """Update crm items by batch requests.

        :param crm_items: list of crm items for update
        """
        queries = []
        for crm_item in crm_items:
            query_params = {
                "entityTypeId": crm_item.fields["entityTypeId"],
                "id": crm_item.fields["id"],
            }
            fields = dict(crm_item.fields)
            fields.pop("id")
            query_params["fields"] = fields
            queries.append(
                BatchQuery(
                    crm_item.fields["id"],
                    crm.Item.UPDATE,
                    query_params,
                ),
            )
        updated_items = {}
        for chunk in build_chunks(queries):
            chunk_json = (await bx24.batch_request(chunk)).json()
            logger.debug(chunk_json)
            if chunk_json["result"]["result_error"]:
                raise BitrixError(str(chunk_json["result"]["result_error"]))
            updated_items |= chunk_json["result"]["result"]
        for crm_item in crm_items:
            crm_item.fields = updated_items[str(crm_item.fields["id"])]["item"]
        return crm_items

    @classmethod
    async def batch_create(
        cls,
        process_id: int,
        crm_items: list[CRMItem],
        as_dict: bool = False,
    ) -> list[CRMItem] | dict[str, CRMItem]:
        """Add crm items by batch requests.

        :param process_id: smart process id in bitrix
        :param crm_items: list of crm items for create
        :param as_dict: result type, defaults to False
        :raises BitrixError: Error in bitrix response
        :return: list or dict of CRMItems
        """
        queries = []
        for crm_item in crm_items:
            query_params = {
                "entityTypeId": process_id,
                "fields": crm_item.fields,
            }
            queries.append(
                BatchQuery(crm_item.name, crm.Item.ADD, query_params),
            )
        created_items: dict | list = {} if as_dict else []
        for chunk in build_chunks(queries):
            chunk_json = (await bx24.batch_request(chunk)).json()
            if chunk_json["result"]["result_error"]:
                raise BitrixError(str(chunk_json["result"]["result_error"]))
            for name, created_item in chunk_json["result"]["result"].items():
                if as_dict:
                    created_items[name] = CRMItem(created_item["item"], name=name)
                else:
                    created_items.append(CRMItem(created_item["item"], name=name))  # type: ignore
        return created_items



"""String query params builders for bx24."""

from dataclasses import dataclass
from urllib.parse import quote

from aiobitrix24.exceptions import BatchError

MAX_BATCH_SIZE = 50
HALT = 0


@dataclass
class BatchQuery:
    """Query for batch request."""

    name: str
    method: str
    params: dict  # noqa: WPS110


def build_query(query_params: dict, prefix: str = "") -> str:
    """Build complex params row from complex dict.

    :param query_params: query params for string conversion
    :param prefix: main param name in deep query, defaults to ""
    :return: specific string with query params for bx24
    """
    query_string = []
    for key, params_value in query_params.items():
        if isinstance(params_value, dict):
            query_string.append(build_query(params_value, key))
        elif isinstance(params_value, list):
            list_params = {
                str(index): parameter for index, parameter in enumerate(params_value)
            }
            query_string.append(build_query(list_params, key))
        else:
            if prefix:
                query_string.append(
                    f"{quote(prefix)}[{quote(str(key))}]={quote(str(params_value))}",
                )
            else:
                query_string.append(f"{quote(str(key))}={quote(str(params_value))}")
    return f"{'&'.join(query_string)}"


def build_batch(queries: list[BatchQuery], halt: int = HALT) -> dict:
    """Build dict for json post batch bitrix24 request.

    :param queries: list of queries for batch request
    :param halt: bitrix24 param, if to stop request on error rises, defaults to HALT
    :raises BatchError: rises if count of queries is more than max batch count
    :return: dict for batch request via json
    """
    if len(queries) > MAX_BATCH_SIZE:
        raise BatchError
    result_query = {"halt": halt, "cmd": {}}
    for query in queries:
        result_query["cmd"][query.name] = f"{query.method}?{build_query(query.params)}"
    return result_query


def build_chunks(
    queries: list[BatchQuery],
    limit: int = MAX_BATCH_SIZE,
) -> list[list[BatchQuery]]:
    """Split list of queries on bathes with limited size.

    :param queries: list of queries for batch request
    :param limit: max count of queries in batch request, defaults to MAX_BATCH_SIZE
    :return: list of batches with queries
    """
    return [
        (queries[key_position : key_position + limit])
        for key_position in range(0, len(queries), limit)
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiobitrix24', 'aiobitrix24.methods']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.0,<0.24.0']

setup_kwargs = {
    'name': 'aiobitrix24',
    'version': '0.4.1',
    'description': 'Simple async bitrix-24 rest client',
    'long_description': '<h2 align="center">AIOBitrix24</h2>\n\n<p align="center">\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>',
    'author': 'andrpocc',
    'author_email': 'andrey.pochatkov@mail.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/prin-it/bitrix24-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
